package utep.plmr.server;

import rice.pastry.Id;
import rice.pastry.Id.Distance;

/**
 * Class for some ID arithmetic that FP does not include
 * or is declared private, like setBit, getBit, absDistance etc.
 * @author derDoc
 *
 */
public class IdHelper {

	/** the length of the bit array in bytes */
	public static int	nlen	= 20;

	/**
	 * set a bit in an ID
	 * @param d
	 * 			the ID
	 * @param i
	 * 			index of bit, from 0..159
	 * 			159 is MSB
	 * @param v
	 * 			the val of new bit, true = 1
	 * @return
	 */
	public static Id setBit(Id d, int i, boolean v) {
		byte[] id = d.copy();
		int index = (i / 8);
		int shift = (i % 8);
		byte val = id[index];
		byte mask = (byte) (1 << shift);

		if (v) {
			id[index] = (byte) (val | mask);
		} else {
			id[index] = (byte) (val & ~mask);
		}
		return Id.build(id);
	}

	/**
	 * Gets a spec. bit of an ID
	 * @param d
	 * 			ID
	 * @param i
	 * 			bits index, 0..159, 159 = MSB
	 * @return
	 * 			the bit as bool, true = 1
	 */
	public static boolean getBit(Id d, int i) {
		byte[] id = d.copy();
		int index = (byte) (i / 8);
		int shift = (byte) (i % 8);
		byte val = id[index];
		byte mask = (byte) (1 << shift);

		if ((val & mask) != 0) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Flips a bit in an ID
	 * @param d
	 * 			ID
	 * @param i
	 * 			the index
	 * @return
	 * 			the new ID
	 */
	public static Id flipBit(Id d, int i) {
		return setBit(d, i, !(getBit(d, i)));
	}

	/**
	 * Absolute numeric distance between two IDs
	 * @param anid
	 * 			ID a
	 * @param bnid
	 * 			ID b
	 * @return
	 * 		the distance as Distance instance
	 */
	public static Distance absDist(Id anid, Id bnid) {
		return new Distance(absDistance(anid, bnid));
	}

	/**
	 * XOR distance between two IDs
	 * @param anid
	 * @param bnid
	 * @return
	 * 			The dist in Distance Dist.
	 */
	public static Distance XORDist(Id anid, Id bnid) {
		return new Distance(XORDistance(anid, bnid));
	}

	/**
	 * XOR distance helper, returns as byte[]
	 * @param anid
	 * @param bnid
	 * @return
	 * 		distance as byte[]
	 */
	public static byte[] XORDistance(Id anid, Id bnid) {
		int x, y, diff;
		byte[] dist = new byte[nlen];
		byte[] aid = anid.copy();
		byte[] bid = bnid.copy();

		for (int i = 0; i < nlen; i++) {
			x = aid[i] & 0x0000000ff;
			y = bid[i] & 0x0000000ff;
			diff = (x ^ y);
			dist[i] = (byte) diff;
		}

		return dist;
	}

	/**
	 * Abs. dist. helper
	 * @param anid
	 * @param bnid
	 * @return
	 * 			abs num dist. as byte[]
	 */
	public static byte[] absDistance(Id anid, Id bnid) {
		int x;
		int y;
		int diff;
		byte carry = 0;
		byte[] dist = new byte[nlen];
		byte[] aid = anid.copy();
		byte[] bid = bnid.copy();

		if (anid.compareTo(bnid) > 0) {
			for (int i = 0; i < nlen; i++) {
				x = aid[i] & 0x0000000ff;
				y = bid[i] & 0x0000000ff;

				diff = x - y - carry;

				if (diff < 0) {
					carry = 1;
				} else {
					carry = 0;
				}

				dist[i] = (byte) diff;
			}
		} else {
			for (int i = 0; i < nlen; i++) {
				x = aid[i] & 0x0000000ff;
				y = bid[i] & 0x0000000ff;

				diff = y - x - carry;

				if (diff < 0) {
					carry = 1;
				} else {
					carry = 0;
				}

				dist[i] = (byte) diff;
			}
		}
		return dist;
	}
	
	/**
	 * Compares two XOR distances
	 * @param da
	 * @param db
	 * @return
	 * 		<0 if a < b
	 * 		>0 if a > b
	 * 		0 if same 
	 */
	public static int XORDistanceComparator(Distance da, Distance db){
		      byte[] ad = da.copy();
		      byte[] bd = db.copy();
		      
		      for (int i = nlen - 1; i >= 0; i--) {
		        if (ad[i] != bd[i]) {
		        	
		          int a = Math.abs(ad[i]);
		          int b = Math.abs(bd[i]);
		        
		          if (a < b) {
		            return -1;
		          } else {
		            return 1;
		          }
		        }
		      }

		      return 0;
	}
}
