package utep.plmr.server;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.util.*;

import rice.Continuation;
import rice.p2p.commonapi.*;
import utep.plmr.CmdExecer;
import utep.plmr.Runners;
import utep.plmr.server.messages.BuildMessage;
import utep.plmr.server.messages.RSyncReqMessage;
import utep.plmr.server.messages.SendResMsg;

/** Class that holds all info about a query for storing in HashTable and later retrieval
 * while sending results and aggregating
 * @author  derDoc
 */
public class Query {

	
	/** time per agg job in ms to wait*/
	public static final int		AGG_WAIT_TIME	= 100;

	/** type of tree this query is using*/
	public int						tree_type;

	public Id						root;

	/** the unique ID of this query per root, used to identify the query at each
	* node*/
	public String					query_ID;

	/** last known parent*/
	public Id						lastParent;

	/** List of byte[]*/
	public ArrayList					results;

	/** List of PathInfo*/
	public ArrayList					paths;

	public String					Agg_cmd;

	/**number of items in path, for easy remembering...*/
	public int						numNodes		= 1;

	/** connection to client, if we are root*/
	public plmr_client_connection	client;
	
	/**If we are waiting for an rsync*/
	public boolean					bAwaitRsync	= false;
	/**and if we need to schedule an init*/
	public String					bAwaitInit_cmd;
	
	/**and children waiting for our rsync to finish, for processing
	*ArrayList<RSyncReqMsg>*/
	public List						childRsyncs = new ArrayList();
	
	/**whether we are collecting path Info data*/
	public boolean						bCollPathInfo;

	public boolean						bFile3in1;

	/**
	 * Creates a query
	 * @param d
	 * 		Root of tree
	 * @param pcc
	 * 			Client Connx that this query came on in, in case we are root
	 * @param p
	 * 			Last known parent, this query came from
	 * @param id
	 * 			ID of query
	 * @param t
	 * 			tree type
	 */
	public Query(Id d, plmr_client_connection pcc, Id p,
			String id, int t) {
		this.results = new ArrayList();
		this.paths = new ArrayList();
		this.root = d;
		this.client = pcc;
		this.lastParent = p;
		this.query_ID = id;
		this.tree_type = t;
	}

	/**
	 * Add a result to this query
	 * @param res
	 * 		the res as byte[]
	 */
	public void addResult(byte[] res) {
		synchronized (this.results) {
			this.results.add(res);
		}
	}

	/**
	 * Returns this instance as a short string
	 */
	public String toString() {
		return new String("Query{" + query_ID + "@" + root + "}");
	}

	/**
	 * Retunrs a longer repr. of this as a String
	 * which includes all infos broker over mult. lines
	 * @return
	 * 			the String
	 */
	public String toFullString() {
		String r = "Query{\n" + 
			"\tQuery_ID:\t" + query_ID + "\n" + 
			"\tRoot:\t" + root + "\n" + 
			"\ttype:\t" + tree_type + "\n" + 
			"\tParent:\t" + lastParent + "\n" + 
			"\tcmd:\t" + Agg_cmd + "\n" +
			"\tbWaitRsync:\t" + bAwaitRsync + "\n" + 
			"\tResults:{\n";

		for (int i = 0; i < results.size(); i++) {
			r += "res[" + i + "]:{\n";
			r += new String((byte[]) results.get(i));
			r += "}\n";
		}
		r += "\t\t}\n}\n";

		return r;
	}
	
	/**
	 * called when sending a res up the tree, to find the parent
	 * checks the type and returns an appr. parent
	 * 
	 * @param q
	 *            The query for info
	 * @return Id The Id of best parent to send res to
	 */
	public Id getNextParent() {

		Id ret = null;

		switch (tree_type) {
			case BuildMessage.CO_KBT:
				ret = lastParent;
				break;
			case BuildMessage.FP_FTT:
				ret = root;
				break;
			case BuildMessage.FP_KBT:
				ret = lastParent;
				break;
		}

		//		System.out.println("getParent() returning:" + ret);
		return ret;
	}
	
	/**
	 * Called to see if the query needs rescheduling of the Aggregator
	 * If it does the aggregator is scheduled for execution
	 * 
	 * @param q
	 *            The query in question
	 * @return int 
	 * 			Vague guess of time needed for completion, 0 iff only one
	 *         result (needs no processing)
	 */
	public int checkAndAgg(Node node) {
		//if we are awaiting rsync just return some time to wait (5secs)
		if (bAwaitRsync)
			return 500 * AGG_WAIT_TIME;

		String cmd = "";
		if (bFile3in1)
			cmd = " " + Runners.S3in1_cmdlines[1];
		final String cmd_l = cmd;
		//needs to be sync'ed on res else we could have a race cond.
		synchronized (results) {

			if (results.size() > 1) {
				System.out.println("Got " + results.size()
						+ " res in queue, sched'ing agg");
				//only if synced
				if (!(Agg_cmd == null || "".equals(Agg_cmd))) {

					((rice.pastry.PastryNode) node).process(
							new rice.Executable() {
								public Object execute() {
									byte[] ls;
									byte[] out = (byte[]) results.remove(0);

									while (out.length < 3) {
										out = (byte[]) results.remove(0);
									}

									ByteArrayOutputStream in = new ByteArrayOutputStream();

									try {
										while (results.size() > 0) {
											ls = (byte[]) results.remove(0);

											//precaution: if len(ls)<3 '()\n': skip
											if (ls.length < 3)
												continue;

											in.reset();

											in.write(String.valueOf(out.length)
													.getBytes());
											in.write('\n');
											in.write(out);
											in.write('\n');

											in.write(String.valueOf(ls.length)
													.getBytes());
											in.write('\n');
											in.write(ls);
											in.write('\n');

											CmdExecer ce = new CmdExecer(
													Agg_cmd + cmd_l,
													in.toByteArray(),
													null);
											ce.run();

											out = ce.stdout;
										}

										results.add(out);
									} catch (Exception e) {
										System.out
												.println("exc in check&agg, agg:exec");
										e.printStackTrace();
										return null;
									}
									return out;
								}
							},
							new Continuation() {
								// will be called if success of the exec
								public void receiveResult(Object result) {}

								// will be called if failure
								public void receiveException(Exception result) {}
							});

					// and signal that caller should wait a little longer by returning
					// >0
					return (results.size() * AGG_WAIT_TIME);

				} else { //just concatenate everything into one big byte array
					//					System.out
					//							.println(this + "-No agg given, concatenating...");
					ByteArrayOutputStream out = new ByteArrayOutputStream();
					byte[] ls;
					try {
						while (results.size() > 0) {
							ls = (byte[]) results.remove(0);
							out.write(ls);
							out.write('\n');
						}
					} catch (IOException ioex) {
						ioex.printStackTrace();
					}

					results.add(out.toByteArray());
					return 0;
				}

			} else if (results.size() == 0) {
				//signal fine... HACK!!, for case when we are not in list
				//sync(this) should take care of case when backg thread is agg'ing
				//i.e. this will block until finished
				return 0;
			} else {
				// signal that everything is fine (only one result)
				return 0;
			}

		}//end sync'ed
	}
	
	/**
	 * Takes a BuildMsg and does the following: 
	 * 	-makes Query Object and adds to known queries HT 
	 * 	-normalizes and/or creates Init/Agg-Execs if contained in msg 
	 * 	-schedules Init for execution 
	 * 	-schedules sending of Res depending on dist(root) 
	 * 	-and sends the msg down the tree It is called on every node until it hits leafs.
	 * 
	 * @param bm
	 *            the BuildMessage
	 * @param pcc
	 *            client_connx that we got this req from if we happen to be the
	 *            root
	 * @param sendDown
	 *            set to true if you want the message to be sent down the Tree
	 *            in case it is new
	 */
	public void checkQuery(BuildMessage bm, plmr_client_connection pcc,
			boolean sendDown, Hashtable queries, Node node, Endpoint endpoint,
			String ssh_username) throws Exception {

		//make new query
		
		if (!queries.containsKey(bm.query_ID)) {

			System.out.println("Building new query for: " + bm);
			String icmd = "";

			Arrays.sort(bm.IPList);
			if ((Arrays.binarySearch(bm.IPList, InetAddress.getLocalHost()
					.getHostAddress()) >= 0)
					|| (Arrays.binarySearch(bm.IPList, InetAddress
							.getLocalHost().getCanonicalHostName()) >= 0)) {

				System.out.println("Scheduling an init run...");

				// create Exec if not already there
				if (bm.Init_cmd == null || bm.Init_cmd.equals("")) {
					if (bm.Init_content != null && bm.Init_content.length > 0)
						icmd = "Init-"
								+ Calendar.getInstance().getTimeInMillis();
				} else
					icmd = bm.Init_cmd;

				if (bm.Init_content != null && bm.Init_content.length > 0) {
					String tstr = new String(bm.Init_content);
					//see if its a hash
					if (tstr.matches("[a-f0-9]{40}")) { //yes!
						//we need to check it, start backgr. processing
						String cmd = plmr_srv_app.HASHER_SCRIPT + " " + icmd;

						byte[] res = (byte[]) new utep.plmr.CmdExecer(
								cmd,
								bm.Init_content,
								null).execute();

						if (new String((byte[]) res).equals("False")) {
							//File outdated request newer
							System.out.println("Hash of: " + icmd
									+ " does not match, req rsync!");

							try {
								bAwaitRsync = true;
								RSyncReqMessage rsm = new RSyncReqMessage(node
										.getId(), bm.from, InetAddress
										.getLocalHost().getHostAddress(), icmd);

								rsm.query_ID = bm.query_ID;
								rsm.username = ssh_username;
								if (client != null)
									client.send(rsm);
								else
									endpoint.route(bm.from, rsm, null);
							} catch (Exception exc) {
								exc.printStackTrace();
							}
						}

						//no hash
					} else {
						File f = new File(icmd);
						if (!"".equals(f.getParentFile())) {
							new File(f.getParentFile().getAbsolutePath())
									.mkdirs();
						}
						FileOutputStream bfw = new FileOutputStream(f);
						bfw.write(bm.Init_content);
						bfw.close();
						icmd = new File(icmd).getAbsolutePath();
						Runtime.getRuntime().exec("chmod +x " + icmd);
						//produce the hash
						String cmd = plmr_srv_app.HASHER_SCRIPT + " ";
						if (bm.RSyncBaseDir != null)
							cmd += bm.RSyncBaseDir;
						else
							cmd += icmd;

						//produce hash in background by calling script
						//System.out.println("Exec: " + cmd);

						((rice.pastry.PastryNode) node).process(
								new utep.plmr.CmdExecer(
										cmd,
										"\n".getBytes(),
										null),
								new Continuation() {
									public void receiveResult(Object res) {}

									public void receiveException(
											Exception result) {}
								});

					}
				}
				final String cmd = icmd;
				String cmd_l = "";
				if (bm.bFile3in1)
					cmd_l = " " + Runners.S3in1_cmdlines[0];

				// schedule producing of result, only if synced
				//if not synced it is deferred until after recpt of syncedMsg
				if (!bAwaitRsync) {
					((rice.pastry.PastryNode) node).process(
							new utep.plmr.CmdExecer(cmd + cmd_l, null, null),
							new Continuation() {
								// will be called if success of the exec
								public void receiveResult(Object result) {
									addResult((byte[]) result);
									//and delete if temp file
									if (cmd.matches("^.*?/Init-(\\d){10,}$")) {
										new File(cmd).delete();
									}
								}

								// will be called if failure
								public void receiveException(Exception result) {
									System.out
											.println("There was an error in Init: "
													+ result);
								}
							});
				} else
					bAwaitInit_cmd = icmd;
			}//end binarySearch = sched init run

			// make cmd for agg

			String gcmd = "";
			if (bm.bFile3in1) {
				gcmd = icmd;

			} else {
				if (bm.Agg_cmd == null || bm.Agg_cmd.equals("")) {
					if (bm.Agg_content != null && bm.Agg_content.length > 0)
						gcmd = "Agg-"
								+ Calendar.getInstance().getTimeInMillis();
				} else
					gcmd = bm.Agg_cmd;
			}
			if (bm.Agg_content != null && bm.Agg_content.length > 0) {
				//see if its a hash
				String tstr = new String(bm.Agg_content);
				if (tstr.matches("[a-f0-9]{40}")) { //yes!
					//we need to check it, start backgr. processing
					String cmd = plmr_srv_app.HASHER_SCRIPT + " ";
					if (bm.RSyncBaseDir != null)
						cmd += bm.RSyncBaseDir;
					else
						cmd += gcmd;

					//exec direct because it would take too long
					byte[] res = (byte[]) new utep.plmr.CmdExecer(
							cmd,
							bm.Agg_content,
							null).execute();

					if (new String(res).equals("False")) {
						//File outdated request never
						System.out.println("Hash of: " + gcmd
								+ " does not match!, requesting rsync");
						try {
							bAwaitRsync = true;
							RSyncReqMessage rsm = new RSyncReqMessage(node
									.getId(), bm.from, InetAddress
									.getLocalHost().getHostAddress(), gcmd);
							rsm.query_ID = bm.query_ID;
							rsm.username = ssh_username;
							if (client != null)
								client.send(rsm);
							else
								endpoint.route(bm.from, rsm, null);
						} catch (Exception exc) {
							exc.printStackTrace();
						}

					}

					//no hash
				} else {

					File f = new File(gcmd);

					if (!"".equals(f.getParent())
							&& !f.getParentFile().isDirectory()) {
						new File(f.getParent()).mkdirs();
					}

					FileOutputStream bfw = new FileOutputStream(f);
					bfw.write(bm.Agg_content);
					bfw.close();
					gcmd = new File(gcmd).getAbsolutePath();
					Runtime.getRuntime().exec("chmod +x " + gcmd);
					//only produce hash in case it is not a dir, then we need to wait till after init
					if (bm.RSyncBaseDir == null) {
						System.out.println("Hashing : " + gcmd);

						((rice.pastry.PastryNode) node).process(
								new utep.plmr.CmdExecer(plmr_srv_app.HASHER_SCRIPT + " "
										+ gcmd, "\n".getBytes(), null),
								new Continuation() {
									public void receiveResult(Object res) {}

									public void receiveException(
											Exception result) {}
								});
					}
				}
			}

			// insert agg into query
			Agg_cmd = gcmd;
			bCollPathInfo = bm.bCollPathInfo;
			bFile3in1 = bm.bFile3in1;
			//  insert query into HT
			queries.put(bm.query_ID, this);

			//set timeout to timeout specified in bm

			long to = bm.timeout;
			/*
			 // schedule sending the ResMsg up, wait for as long as distance to
			 // root.
			 //get the absolute distance to the root, will be 0 if we are root
			 byte[] b = IdHelper.XORDistance(
			 (rice.pastry.Id) bm.root,
			 (rice.pastry.Id) node.getId());
			 //the timeout
			 long to = 0l;
			 int ti;
			 //and for every byte, sub from 256, shl a lil' (0 is idx of LS byte)
			 //and mult by wait_time
			 //add that all up
			 for (int i = 0; i < 20; i++) {
			 //since its byte it might be negative, check and invert if so
			 ti = ((b[i] < 0) ? (b[i] * (-1)) : (b[i]));
			 to += ((0xff - ti) << (i)) * WAIT_TIME;
			 }

			 long tof = Math.round( to / WAIT_DIV_TIME);
			 to = tof;
			 */

			System.out
					.println("and scheduling sending of res in " + to + " ms");

			// and schedule a SendResMsg to us in 'to' millis
			((rice.pastry.commonapi.PastryEndpoint) endpoint).scheduleMessage(
					new SendResMsg(this, false),
					to);

		}//end unknwon query
		else {
			System.out.println(this + " got known BM for query "
					+ queries.get(bm.query_ID));
			sendDown = false;
		}

		if (sendDown)
			bm.sendDown(node, endpoint);
	}

}
